/*
 * Copyright (c) 2018 naehrwert
 * Copyright (c) 2018-2023 CTCaer
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <storage/sd.h>
#include <storage/sdmmc.h>
#include <storage/sdmmc_driver.h>
#include <gfx_utils.h>
#include <libs/fatfs/ff.h>
#include <mem/heap.h>

#ifndef BDK_SDMMC_UHS_DDR200_SUPPORT
#define SD_DEFAULT_SPEED SD_UHS_SDR104
#else
#define SD_DEFAULT_SPEED SD_UHS_DDR208
#endif

static bool sd_mounted = false;
static bool sd_init_done = false;
static bool insertion_event = false;
static u16  sd_errors[3] = { 0 }; // Init and Read/Write errors.
static u32  sd_mode = SD_DEFAULT_SPEED;


void sd_error_count_increment(u8 type)
{
	emmc_error_count_increment(type);
}

u16 *sd_get_error_count()
{
	return emmc_get_error_count();
}

bool sd_get_card_removed()
{
	// if (insertion_event && !sdmmc_get_sd_inserted())
	// 	return true;

	// you probably wouldn't want to remove the nand while the system is powered on anyways.
	return false;
}

bool sd_get_card_initialized()
{
	// return sd_init_done;
	return true;
}

bool sd_get_card_mounted()
{
	return sd_mounted;
}

u32 sd_get_mode()
{
	return sd_mode;
}

int sd_init_retry(bool power_cycle)
{
	return emmc_init_retry(power_cycle);
}

bool sd_initialize(bool power_cycle)
{
	return emmc_initialize(power_cycle);
}

bool sd_mount()
{
	if (sd_init_done && sd_mounted)
		return true;

	int res = 0;

	if (!sd_init_done)
		res = !sd_initialize(false);

	if (res)
	{
		gfx_con.mute = false;
		EPRINTF("Failed to init SD card.");
		if (!sdmmc_get_sd_inserted())
			EPRINTF("Make sure that it is inserted.");
		else
			EPRINTF("SD Card Reader is not properly seated!");
	}
	else
	{
		if (!sd_mounted)
			res = f_mount(&sd_fs, "0:", 1); // Volume 0 is SD.
		if (res == FR_OK)
		{
			sd_mounted = true;
			return true;
		}
		else
		{
			gfx_con.mute = false;
			EPRINTFARGS("Failed to mount SD card (FatFS Error %d).\nMake sure that a FAT partition exists..", res);
		}
	}

	return false;
}

static void _sd_deinit(bool deinit)
{
	if (deinit)
	{
		insertion_event = false;
		if (sd_mode == SD_INIT_FAIL)
			sd_mode = SD_DEFAULT_SPEED;
	}

	if (sd_init_done)
	{
		if (sd_mounted)
			f_mount(NULL, "0:", 1); // Volume 0 is SD.

		if (deinit)
		{
			sdmmc_storage_end(&sd_storage);
			sd_init_done = false;
		}
	}
	sd_mounted = false;
}

void sd_unmount() { _sd_deinit(false); }
void sd_end()     { emmc_end(); }

bool sd_is_gpt()
{
	return sd_fs.part_type;
}

void *sd_file_read(const char *path, u32 *fsize)
{
	FIL fp;
	if (!sd_get_card_mounted())
		return NULL;

	if (f_open(&fp, path, FA_READ) != FR_OK)
		return NULL;

	u32 size = f_size(&fp);
	if (fsize)
		*fsize = size;

	void *buf = malloc(size);

	if (f_read(&fp, buf, size, NULL) != FR_OK)
	{
		free(buf);
		f_close(&fp);

		return NULL;
	}

	f_close(&fp);

	return buf;
}

int sd_save_to_file(void *buf, u32 size, const char *filename)
{
	FIL fp;
	u32 res = 0;
	if (!sd_get_card_mounted())
		return FR_DISK_ERR;

	res = f_open(&fp, filename, FA_CREATE_ALWAYS | FA_WRITE);
	if (res)
	{
		EPRINTFARGS("Error (%d) creating file\n%s.\n", res, filename);
		return res;
	}

	f_write(&fp, buf, size, NULL);
	f_close(&fp);

	return 0;
}
